/*
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "include/compiler.h"
#include "include/physical_projection.h"
#include "include/physical_get.h"
#include "include/physical_group_by.h"
#include "include/physical_predicate.h"
#include "include/physical_join.h"

#include <iostream>
#include <sstream>

namespace ppds {

    std::shared_ptr<PhysicalOperator> CompilePhysicalPlanRecursive(const PlanNode& logical_plan) {
        std::shared_ptr<PhysicalOperator> physical_op;
        switch (logical_plan.type) {
            case NodeType::TableScan: {
                physical_op = std::make_shared<PhysicalGet>(logical_plan.parameters.at("file_path"));
                break;
            }
            case NodeType::Projection: {
                std::vector<uint16_t> column_ids;
                const std::string& column_ids_unparsed = logical_plan.parameters.at("column_ids");

                std::string token;
                for (const auto c : column_ids_unparsed) {
                    if (c == ',') {
                        column_ids.push_back(std::stoi(token));
                        token.clear();
                        continue;
                    }
                    token += c;
                }
                column_ids.push_back(std::stoi(token));

                physical_op = std::make_shared<PhysicalProjection>(
                        CompilePhysicalPlanRecursive(logical_plan.children[0]),
                        std::move(column_ids));
                break;
            }
            case NodeType::Predicate: {
                uint16_t column_id = std::stoi(logical_plan.parameters.at("column_id"));
                const std::string& cond_str = logical_plan.parameters.at("condition");
                Condition condition;
                if (cond_str == "=") {
                    condition = Condition::Equals;
                } else if (cond_str == "!=") {
                    condition = Condition::NotEquals;
                } else if (cond_str == "<") {
                    condition = Condition::LowerThan;
                } else if (cond_str == "<=") {
                    condition = Condition::LowerThanEquals;
                } else if (cond_str == ">") {
                    condition = Condition::GreaterThan;
                } else if (cond_str == ">=") {
                    condition = Condition::GreaterThanEquals;
                } else {
                    throw std::logic_error("Unknown condition: " + cond_str);
                }

                const std::string& literal_str = logical_plan.parameters.at("literal");
                std::variant<int32_t, float, std::string> literal;
                char* int_end;
                strtol(literal_str.c_str(), &int_end, 10);
                if (!*int_end) {
                    literal = std::stoi(literal_str);
                } else {
                    char* float_end;
                    strtof(literal_str.c_str(), &float_end);
                    if (!*float_end) {
                        literal = std::stof(literal_str);
                    } else {
                        literal = literal_str;
                    }
                }


                physical_op = std::make_shared<PhysicalPredicate>(
                        CompilePhysicalPlanRecursive(logical_plan.children[0]),
                        column_id, condition,literal);
                break;
            }
            case NodeType::EquiJoin: {
                uint16_t left_column_id = std::stoi(logical_plan.parameters.at("left_column_id"));
                uint16_t right_column_id = std::stoi(logical_plan.parameters.at("right_column_id"));
                physical_op = std::make_shared<PhysicalJoin>(CompilePhysicalPlanRecursive(logical_plan.children[0]),
                                                             CompilePhysicalPlanRecursive(logical_plan.children[1]),
                                                             left_column_id, right_column_id);
                break;
            }
            case NodeType::GroupBy: {
                std::optional<uint16_t> group_by_column_id = std::nullopt;
                if (logical_plan.parameters.find("group_by_column_id") != logical_plan.parameters.end()) {
                    group_by_column_id = std::stoi(logical_plan.parameters.at("group_by_column_id"));
                }

                uint16_t aggregate_column_id = std::stoi(logical_plan.parameters.at("aggregate_column_id"));
                AggregateType aggregate_type;
                const std::string& agg_type_str = logical_plan.parameters.at("aggregate_type");
                if (agg_type_str == "avg") {
                    aggregate_type = AggregateType::Average;
                } else if (agg_type_str == "max") {
                    aggregate_type = AggregateType::Maximum;
                } else if (agg_type_str == "min") {
                    aggregate_type = AggregateType::Minimum;
                } else {
                    throw std::logic_error("Unknown aggregate type: " + agg_type_str);
                }
                physical_op = std::make_shared<PhysicalGroupBy>(
                        CompilePhysicalPlanRecursive(logical_plan.children[0]),
                                                                group_by_column_id,
                                                                aggregate_type,
                                                                aggregate_column_id);
            }
        }
        return physical_op;
    }

    std::shared_ptr<ONCIterator> Compiler::CompilePlan(const PlanNode& node) {
        return CompilePhysicalPlanRecursive(node);
    }

}
