/*
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#pragma once

#include "physical_operator.h"
#include "types.h"

#include <fstream>

namespace ppds {
    constexpr char SEPARATOR = '|';

    class GetOperatorState : public ONCIteratorState {
    public:
        std::ifstream file;
        std::shared_ptr<std::vector<DataType>> data_types;
        std::vector<std::function<void(std::string&,Record&,uint16_t)>> parse_funcs;
        bool scan_done = false;
    };

    class PhysicalGet : public PhysicalOperator {
    public:
        PhysicalGet(std::string file_dir_p) : file_dir(std::move(file_dir_p)) {}

        static std::function<void(std::string&,Record&,uint16_t)> CreateParseFunction(DataType type) {
            switch (type) {
                case DataType::Integer:
                    return [](std::string& input ,Record& rec, uint16_t index) {
                        int32_t value = std::stoi(input);
                        rec.SetValue(index, value);
                    };
                case DataType::Float:
                    return [](std::string& input ,Record& rec, uint16_t index) {
                        float value = std::stof(input);
                        rec.SetValue(index, value);
                    };
                case DataType::String:
                    return [](std::string& input ,Record& rec, uint16_t index) {
                        rec.SetValue(index, input);
                    };
            }
        }

        void Open() override {
            auto get_state = std::make_shared<GetOperatorState>();
            state = get_state;

            get_state->file = std::ifstream(file_dir);
            get_state->data_types = std::make_shared<std::vector<DataType>>();
            // Guess types
            std::string line;
            if (!std::getline(get_state->file, line)) {
                throw std::logic_error("Table file is empty.");
            }

            std::string token;
            for (char c : line) {
                if (c == SEPARATOR) {
                    char* int_end;
                    strtol(token.c_str(), &int_end, 10);
                    if (!*int_end) {
                        get_state->data_types->push_back(DataType::Integer);
                        get_state->parse_funcs.push_back(CreateParseFunction(DataType::Integer));
                        token.clear();
                        continue;
                    }
                    char* float_end;
                    strtof(token.c_str(), &float_end);
                    if (!*float_end) {
                        get_state->data_types->push_back(DataType::Float);
                        get_state->parse_funcs.push_back(CreateParseFunction(DataType::Float));
                        token.clear();
                        continue;
                    }
                    get_state->data_types->push_back(DataType::String);
                    get_state->parse_funcs.push_back(CreateParseFunction(DataType::String));
                    token.clear();
                    continue;
                }
                token += c;
            }

            get_state->file.seekg(0);
        };

        void Next(Record& result) override {
            auto& get_state = (GetOperatorState&) *state;

            if (get_state.scan_done) {
                result.Clear();
                return;
            }

            std::string line;
            if (!std::getline(get_state.file, line)) {
                get_state.scan_done = true;
                result.Clear();
                return;
            }

            if (!result.IsValid()) {
                result.Initialize(get_state.data_types);
            }

            uint16_t column_index = 0;
            std::string token;
            for (char c : line) {
                if (c == SEPARATOR) {
                    get_state.parse_funcs[column_index](token, result, column_index);
                    token.clear();
                    column_index++;
                    continue;
                }
                token += c;
            }

            // If line does not end with separator
            if (!token.empty()) {
                get_state.parse_funcs[column_index](token, result, column_index);
            }
        }

        void Close() override {
            auto& get_state = (GetOperatorState&) *state;
            get_state.file.close();
            state = nullptr;
        }

    private:
        const std::string file_dir;
    };
}
