/*
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#pragma once

#include "hash_table.h"
#include "physical_operator.h"
#include "types.h"

namespace ppds {
    class JoinOperatorState : public ONCIteratorState {
    public:
        Record left_child_output;
        Record right_child_output;
        bool build_phase_done = false;
        std::unique_ptr<HashTable> ht;
        std::shared_ptr<std::vector<DataType>> build_types;
        std::function<void(JoinOperatorState&,Record&)> probe_func;
        std::vector<Record> left_overs;
    };

    class PhysicalJoin : public PhysicalOperator {
    public:
        PhysicalJoin(std::shared_ptr<PhysicalOperator> left_child, std::shared_ptr<PhysicalOperator> right_child,
                     uint16_t left_join_column_id_p, uint16_t right_join_column_id_p) :
                left_join_column_id(left_join_column_id_p), right_join_column_id(right_join_column_id_p) {
            children.reserve(2);
            children.push_back(std::move(left_child));
            children.push_back(std::move(right_child));
        }

        void Open() override {
            auto join_state = std::make_shared<JoinOperatorState>();
            state = join_state;
            children[0]->Open();
            children[1]->Open();
        }

        template<class T>
        void FillHashTable(JoinOperatorState& state) const {
            auto& ht = (HashTableImpl<T>&) *state.ht;

            while (state.right_child_output.IsValid()) {
                ht.map.emplace(std::get<T>(state.right_child_output.GetValue(right_join_column_id)),
                               state.right_child_output);
                children[1]->Next(state.right_child_output);
            }
        }

        template<class T>
        std::function<void(JoinOperatorState&,Record&)> GenerateProbeFunction() const{
            return [&](JoinOperatorState& s, Record& r) {
                auto& ht = (HashTableImpl<T>&) *s.ht;
                children[0]->Next(s.left_child_output);

                if (!r.IsValid() && s.left_child_output.IsValid()) {
                    auto data_types = std::make_shared<std::vector<DataType>>();
                    auto left_column_types = s.left_child_output.GetColumnTypes();
                    data_types->insert(data_types->end(), left_column_types->begin(), left_column_types->end());
                    data_types->insert(data_types->end(), s.build_types->begin(), s.build_types->end());
                    r.Initialize(data_types);
                }

                while (s.left_child_output.IsValid()) {
                    auto& val = std::get<T>(s.left_child_output.GetValue(left_join_column_id));

                    uint64_t match_count = 0;
                    auto eq_range = ht.map.equal_range(val);
                    for (auto it = eq_range.first; it != eq_range.second; ++it) {
                        match_count++;
                        auto& lhs = s.left_child_output;
                        auto& rhs = it->second;

                        if (match_count == 1) {
                            for (uint16_t i = 0; i< lhs.GetColumnCount(); i++) {
                                r.SetValue(i, lhs.GetValue(i));
                            }
                            for (uint16_t i = 0; i< rhs.GetColumnCount(); i++) {
                                r.SetValue(i + lhs.GetColumnCount(), rhs.GetValue(i));
                            }
                        } else {
                            Record left_over;
                            left_over.Initialize(r.GetColumnTypes());
                            for (uint16_t i = 0; i< lhs.GetColumnCount(); i++) {
                                left_over.SetValue(i, lhs.GetValue(i));
                            }
                            for (uint16_t i = 0; i< rhs.GetColumnCount(); i++) {
                                left_over.SetValue(i + lhs.GetColumnCount(), rhs.GetValue(i));
                            }
                            s.left_overs.push_back(std::move(left_over));
                        }
                    }
                    if (match_count == 0) {
                        children[0]->Next(s.left_child_output);
                        continue;
                    } else if (match_count == 1) {
                        return;
                    } else {
                        std::reverse(s.left_overs.begin(), s.left_overs.end());
                        return;
                    }

                }
                r.Clear();
            };
        }

        void Next(Record& result) override {
            auto& join_state = (JoinOperatorState&) *state;

            // Build
            if (!join_state.build_phase_done) {
                children[1]->Next(join_state.right_child_output);
                if (!join_state.right_child_output.IsValid()) {
                    result.Clear();
                    return;
                }

                join_state.build_types = join_state.right_child_output.GetColumnTypes();

                switch (join_state.right_child_output.GetColumnType(right_join_column_id)) {
                    case DataType::Integer: {
                        join_state.ht = std::make_unique<HashTableImpl<int32_t>>();
                        FillHashTable<int32_t>(join_state);
                        join_state.probe_func = GenerateProbeFunction<int32_t>();
                        break;
                    }
                    case DataType::Float: {
                        join_state.ht = std::make_unique<HashTableImpl<float>>();
                        FillHashTable<float>(join_state);
                        join_state.probe_func = GenerateProbeFunction<float>();
                        break;
                    }
                    case DataType::String: {
                        join_state.ht = std::make_unique<HashTableImpl<std::string>>();
                        FillHashTable<std::string>(join_state);
                        join_state.probe_func = GenerateProbeFunction<std::string>();
                        break;
                    }
                }
                join_state.build_phase_done = true;
            }

            if (!join_state.left_overs.empty()) {
                result = join_state.left_overs.back();
                join_state.left_overs.pop_back();
                return;
            }

            // Probe
            join_state.probe_func(join_state, result);
        }

        void Close() override {
            auto& join_state = (JoinOperatorState&) state;
            children[1]->Close();
            children[0]->Close();
            state = nullptr;
        }

    private:
        const uint16_t left_join_column_id;
        const uint16_t right_join_column_id;
    };

}
