/*
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#pragma once

#include "physical_operator.h"
#include "types.h"

#include <algorithm>

namespace ppds {
    class PredicateOperatorState : public ONCIteratorState {
    public:
        Record child_output;
        std::function<bool(Record&)> predicate_func;
    };

    enum class Condition {
        Equals,
        NotEquals,
        LowerThan,
        LowerThanEquals,
        GreaterThan,
        GreaterThanEquals
    };

    class PhysicalPredicate : public PhysicalOperator {
    public:
        PhysicalPredicate(std::shared_ptr<PhysicalOperator> child, uint16_t column_id_p, Condition condition_p,
                          std::variant<int32_t, float, std::string> literal_p) :
                          column_id(column_id_p), condition(condition_p), literal(std::move(literal_p)) {
            children.push_back(std::move(child));
        }

        template<class T>
        static std::function<bool(Record&)> GenerateExpression(uint16_t column_id, Condition condition, T literal) {
            std::function<bool(Record&)> func;
            switch (condition) {
                case Condition::Equals: {
                    func = [column_id,literal] (Record& r) {return std::get<T>(r.GetValue(column_id)) == literal;}; break;
                }
                case Condition::NotEquals: {
                    func = [column_id,literal] (Record& r) {return std::get<T>(r.GetValue(column_id)) != literal;}; break;
                }
                case Condition::LowerThan: {
                    func = [column_id,literal] (Record& r) {return std::get<T>(r.GetValue(column_id)) < literal;}; break;
                }
                case Condition::LowerThanEquals: {
                    func = [column_id,literal] (Record& r) {return std::get<T>(r.GetValue(column_id)) <= literal;}; break;
                }
                case Condition::GreaterThan: {
                    func = [column_id,literal] (Record& r) {return std::get<T>(r.GetValue(column_id)) > literal;}; break;
                }
                case Condition::GreaterThanEquals: {
                    func = [column_id,literal] (Record& r) {return std::get<T>(r.GetValue(column_id)) >= literal;}; break;
                }
            }
            return func;
        }

        void Open() override {
            auto pred_state = std::make_shared<PredicateOperatorState>();
            state = pred_state;
            children[0]->Open();
            if (std::holds_alternative<int32_t>(literal)) {
                pred_state->predicate_func = GenerateExpression(column_id, condition, std::get<int32_t>(literal));
            } else if (std::holds_alternative<float>(literal)) {
                pred_state->predicate_func = GenerateExpression(column_id, condition, std::get<float>(literal));
            } else if (std::holds_alternative<std::string>(literal)) {
                if (condition != Condition::Equals && condition != Condition::NotEquals) {
                    throw std::logic_error("Operator only supports (n)eq string comparison.");
                }
                pred_state->predicate_func = GenerateExpression(column_id, condition, std::get<std::string>(literal));
            }
        };

        void Next(Record& result) override {
            auto& pred_state = (PredicateOperatorState&) *state;

            children[0]->Next(pred_state.child_output);
            if (!pred_state.child_output.IsValid()) {
                result.Clear();
                return;
            }

            if (!result.IsValid()) {
                result.Initialize(pred_state.child_output.GetColumnTypes());
            }

            while (pred_state.child_output.IsValid()) {
                if (pred_state.predicate_func(pred_state.child_output)) {
                    // Copy record
                    for (uint16_t i = 0; i < result.GetColumnCount(); i++) {
                        result.SetValue(i, pred_state.child_output.GetValue(i));
                    }
                    return;
                }
                children[0]->Next(pred_state.child_output);
            }
            result.Clear();
        }

        void Close() override {
            state = nullptr;
            children[0]->Close();
        }

    private:
        const uint16_t column_id;
        const Condition condition;
        const std::variant<int32_t, float, std::string> literal;
    };
}
